#!/usr/bin/env python3
# coding=utf-8
#
# The Fairy's Throne -- a semi-graphical roguelike
# Copyright 2018 Felix Pleșoianu <https://felix.plesoianu.ro/>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import division
from __future__ import print_function

import random
import sys
import time

if sys.version_info.major >= 3:
	from tkinter import *
	from tkinter import ttk
	from tkinter.messagebox import showinfo, askyesno
	from tkinter.simpledialog import askinteger
else:
	from Tkinter import *
	import ttk
	from tkMessageBox import showinfo, askyesno
	from tkSimpleDialog import askinteger

blurb = "The Fairy King has promised the granting of a wish to any mortal who can reach them in their throne room. But none who entered the palace ever came back. Now it's your turn to challenge..."

about_text = """The Fairy's Throne
a 2018 game by No Time To Play

Available under the MIT License.
See the source code for details."""

help_text = """The Fairy's Throne can be played entirely with the GUI controls.
Also, apart from the menu shortcuts, you can use these keys:

?	- show this help text;
@	- show the character sheet;
i	- show the inventory window;
g	- get an item from the floor;
m	- cast magical spell;

arrows	- move around/attack;
h,j,k,l	- same as above;
>	- climb to the next floor if on stairs;
<	- climb to the previous floor if on stairs;
space	- contextual action: get an item or climb stairs, if any;

escape	- close charsheet/inventory/help window;
F10	- open game menu;
menu	- same as above.

The following keys only work in the inventory window:

w	- wield or wear item;
e	- eat item;
d	- drop item.

You can also click in the viewport to examine a tile's content."""

scale = 24
hud_font = "helvetica 18 bold"
map_font = "courier 24 bold"
crimson = "#DC143C" # Tkinter on Mac OS X doesn't support this color name.
lime = "#32CD32" # Or this one.

quiet = False

top = Tk()
top.title("The Fairy's Throne")
top.resizable(FALSE, FALSE)

#top.bind("<Key-Escape>", lambda event: top.iconify())

viewport = Canvas(top, width=40 * scale, height=25 * scale, background="black")
viewport.pack(side=TOP)

toolbar = ttk.Frame(top)
toolbar.pack(side=BOTTOM, fill="x", padx=4, pady=4)

console = Text(
	top, background="black", foreground="white",
	font=hud_font, width=40, height=2)
console.pack(side=LEFT, fill="x", expand=TRUE)

scroll = ttk.Scrollbar(
	top, orient=VERTICAL, command=console.yview)
console["yscrollcommand"] = scroll.set
scroll.pack(side=RIGHT, fill="y")

def render_title_screen(canvas):
	canvas.create_text(
		480, 150, text=blurb, font="Times 18 bold", fill="indianred",
		width=600, justify="center", tags="title")
	canvas.create_text(
		480, 300, text="The Fairy's Throne", font="Times 72 bold",
		fill="gold", justify="center", tags="title")
	canvas.create_text(
		480, 450, text="a game by\nNo Time To Play",
		font="Times 36 bold", fill="turquoise",
		justify="center", tags="title")

def log_to_console(message):
	console.insert("end", "\n")
	console.insert("end", message)
	console.yview_moveto(1)
	
	if not quiet: print(message)

def confirmed_quit(window):
	answer = askyesno(
		parent=window, message="Quit The Fairy's Throne?",
		title="Are you sure?", icon="question")
	if answer:
		window.destroy()

class HUD:
	def __init__(self, canvas):
		self.canvas = canvas
		self.slots = {}
		self.slots["nw"] = self.canvas.create_text(
			0, 0, fill="white",
			anchor="nw", justify="left",
			font=hud_font, tags="hud")
		self.slots["n"] = self.canvas.create_text(
			480, 0, fill="white",
			anchor="n", justify="center",
			font=hud_font, tags="hud")
		self.slots["ne"] = self.canvas.create_text(
			960, 0, fill="white",
			anchor="ne", justify="right",
			font=hud_font, tags="hud")
		self.slots["sw"] = self.canvas.create_text(
			0, 600, fill="white",
			anchor="sw", justify="left",
			font=hud_font, tags="hud")
		self.slots["s"] = self.canvas.create_text(
			480, 600, fill="white",
			anchor="s", justify="center",
			font=hud_font, tags="hud")
		self.slots["se"] = self.canvas.create_text(
			960, 600, fill="white",
			anchor="se", justify="right",
			font=hud_font, tags="hud")

	def __setitem__(self, slot, msg):
		if slot in self.slots:
			self.canvas.itemconfigure(self.slots[slot], text=msg)
		else:
			raise KeyError("No such slot: " + slot)
		
	def lift(self):
		for i in self.slots:
			self.canvas.lift(self.slots[i])
	
	def clear(self):
		for i in self.slots:
			self.canvas.delete(self.slots[i])

class MapView:
	def __init__(self, canvas):
		self.canvas = canvas

	def make_tile_icon(self, char, x, y):
		x1 = x * scale
		y1 = y * scale
		x2 = x1 + scale
		y2 = y1 + scale

		if char == '.':
			return 0
		if char == '#':
			return self.canvas.create_rectangle(
				x1, y1, x2, y2,
				fill="firebrick", disabledfill="#561111",
				tags="map")
		elif char == '|':
			return self.canvas.create_rectangle(
				x1 + scale // 3,
				y1 + scale // 8,
				x2 - scale // 3,
				y2 - scale // 8,
				fill="gray50", disabledfill="gray25",
				tags="map")
		elif char == '*':
			return self.canvas.create_oval(
				x1 + scale // 8,
				y1 + scale // 8,
				x2 - scale // 8,
				y2 - scale // 8,
				width=4, fill="turquoise", outline="gray50",
				disabledfill="turquoise4",
				disabledoutline="gray25",
				tags="map")
		elif char == '&':
			return self.canvas.create_text(
				x1 + scale // 2,
				y1 + scale // 2,
				text='&', font=map_font,
				fill="orangered", disabledfill="orangered4",
				tags="map")
		elif char == '"':
			return self.canvas.create_text(
				x1 + scale // 2,
				y1 + scale // 2,
				text='"', font=map_font,
				fill="springgreen3",
				disabledfill="springgreen4",
				tags="map")
		elif char == '=':
			return self.canvas.create_text(
				x1 + scale // 2,
				y1 + scale // 2,
				text='=', font=map_font,
				fill="seagreen",
				disabledfill="darkolivegreen",
				tags="map")
		elif char == '<':
			return self.canvas.create_line(
				x1, y1,
				x1 + scale // 3, y1,
				x1 + scale // 3, y1 + scale // 3,
				x2 - scale // 3, y1 + scale // 3,
				x2 - scale // 3, y2 - scale // 3,
				x2, y2 - scale // 3,
				x2, y2,
				width=4, fill="gray50", disabledfill="gray25",
				tags="map")
		elif char == '>':
			return self.canvas.create_line(
				x1, y2,
				x1, y2 - scale // 3,
				x1 + scale // 3, y2 - scale // 3,
				x1 + scale // 3, y1 + scale // 3,
				x2 - scale // 3, y1 + scale // 3,
				x2 - scale // 3, y1,
				x2, y1,
				width=4,
				fill="gray50", disabledfill="gray25",
				tags="map")
		elif char == '+':
			return self.canvas.create_text(
				x1 + scale // 2,
				y1 + scale // 2,
				text='+', font=map_font,
				fill="gold", disabledfill="goldenrod",
				tags="map")
		else:
			return 0

	def make_thing_icon(self, char, x, y, color):
		return self.canvas.create_text(
			x * scale + scale // 2,
			y * scale + scale // 2,
			text=char, font=map_font,
			fill=color, tags="map thing")

	def hide_tile(self, tile):
		self.canvas.itemconfigure(tile, state="hidden")

	def show_tile(self, tile):
		self.canvas.itemconfigure(tile, state="normal")

	def fade_tile(self, tile):
		self.canvas.itemconfigure(tile, state="disabled")

	def clear(self):
		self.canvas.delete("map")

helpdlg = Toplevel(top)
helpdlg.transient(top)
helpdlg.title("How to play")
helpdlg.protocol("WM_DELETE_WINDOW", helpdlg.withdraw)
helpdlg.bind("<Key-Escape>", lambda event: helpdlg.withdraw())
helpdlg.resizable(FALSE, FALSE)
helpdlg.withdraw()

top.bind("<Key-question>", lambda event: helpdlg.deiconify())
top.bind("<Key-F1>", lambda event: helpdlg.deiconify())

helpmsg = ttk.Label(helpdlg, text=help_text, relief="groove")
helpmsg.pack(padx=4, pady=4, ipadx=4, ipady=4)

game_menu = Menu(top, tearoff=0)
game_menu.add_command(label="New...", underline=0, accelerator="Ctrl-N")
game_menu.add_command(
	label="Clear log", underline=0,
	command=lambda:console.delete('1.0', 'end'))
game_menu.add_separator()
game_menu.add_command(
	label="Help", underline=0, accelerator="F1",
	command=lambda:helpdlg.deiconify())
game_menu.add_command(label="About", underline=0,
	command=lambda:showinfo("About", about_text, parent=top))
#game_menu.add_command(
#	label="High scores", underline=5, state="disabled")
game_menu.add_separator()
game_menu.add_command(
	label="Quit", underline=0, accelerator="Ctrl-Q",
	command=lambda: confirmed_quit(top))

menu_button = ttk.Menubutton(
	toolbar, text="Game", underline=0, menu=game_menu, direction="above")
menu_button.pack(side=LEFT)

top.bind("<Key-F10>", lambda event: menu_button.event_generate("<<Invoke>>"))
top.bind("<Key-Menu>", lambda event: menu_button.event_generate("<<Invoke>>"))
top.bind("<Control-q>", lambda event: confirmed_quit(top))
top.protocol("WM_DELETE_WINDOW", lambda: confirmed_quit(top))

all_buttons = {}

cmdkeys = ["Stats", "Items", "Take", "Magic"]
for lbl in cmdkeys:
	button = ttk.Button(toolbar, text=lbl, underline=0)
	button.pack(side=LEFT)
	all_buttons[lbl] = button

dirkeys = ["Climb", "Right", "Up", "Down", "Left"]
for lbl in dirkeys:
	button = ttk.Button(toolbar, text=lbl, underline=0)
	button.pack(side=RIGHT)
	all_buttons[lbl] = button

charsheet = Toplevel(top)
charsheet.transient(top)
charsheet.title("Character sheet")
charsheet.protocol("WM_DELETE_WINDOW", charsheet.withdraw)
charsheet.bind("<Key-Escape>", lambda event: charsheet.withdraw())
charsheet.resizable(FALSE, FALSE)
charsheet.withdraw()

all_buttons["Stats"].configure(command=charsheet.deiconify)
top.bind("<Key-at>", lambda event: charsheet.deiconify())

attr_frame = ttk.Labelframe(charsheet, text="Attributes", underline=1)
attr_frame.grid(row=0, column=0, rowspan=3, padx=4, pady=4, sticky="ew")
attrs = ["muscle", "agility", "stamina"]
attr_display = {}
attr_buttons = {}
for i in range(len(attrs)):
	field = attrs[i]
	lbl = ttk.Label(attr_frame, text=field.capitalize(), underline=0)
	lbl.grid(row=i, column=0, sticky="w", padx=4, pady=4)
	attr_display[field] = ttk.Label(attr_frame, text="2")
	attr_display[field].grid(row=i, column=1, sticky="e", padx=4, pady=4)
	attr_frame.grid_columnconfigure(attr_display[field], weight=1)
	attr_buttons[field] = ttk.Button(
		attr_frame, text="+1", width=2, state="disabled")
	attr_buttons[field].grid(row=i, column=2, padx=4, pady=4)

xp_frame = ttk.Labelframe(charsheet, text="Experience", underline=1)
xp_frame.grid(row=0, column=1, padx=4, pady=4, sticky="ew")
xp_fields = ["Total", "Unused"]
xp_display = {}
for i in range(len(xp_fields)):
	field = xp_fields[i]
	lbl = ttk.Label(xp_frame, text=field)
	lbl.grid(row=i, column=0, sticky="w", padx=4, pady=4)
	xp_display[field] = ttk.Label(xp_frame, text="0")
	xp_display[field].grid(row=i, column=1, sticky="e", padx=4, pady=4)

lbl_wounds = ttk.Label(charsheet, text="Wounds: 0/0")
lbl_wounds.grid(row=1, column=1, padx=4, pady=4)
lbl_fortune = ttk.Label(charsheet, text="")
lbl_fortune.grid(row=2, column=1, padx=4, pady=4)

gear_frame = ttk.Labelframe(charsheet, text="Equipment", underline=0)
gear_frame.grid(row=3, column=0, padx=4, pady=4, sticky="ew")
slots = ["Weapon", "Armor", "Spell"]
gear_display = {}
for i in range(len(slots)):
	field = slots[i]
	lbl = ttk.Label(gear_frame, text=field)
	lbl.grid(row=i, column=0, sticky="w", padx=4, pady=4)
	gear_display[field] = ttk.Label(gear_frame, text="None")
	gear_display[field].grid(row=i, column=1, sticky="w", padx=4, pady=4)

bonus_frame = ttk.Labelframe(charsheet, text="Combat", underline=0)
bonus_frame.grid(row=3, column=1, padx=4, pady=4, sticky="ew")
bonus_fields = ["Attack", "Defense", "Damage"]
bonus_display = {}
for i in range(len(bonus_fields)):
	field = bonus_fields[i]
	lbl = ttk.Label(bonus_frame, text=field)
	lbl.grid(row=i, column=0, sticky="w", padx=4, pady=4)
	bonus_display[field] = ttk.Label(bonus_frame, text="0")
	bonus_display[field].grid(row=i, column=1, sticky="e", padx=4, pady=4)

inventory = Toplevel(top)
inventory.transient(top)
inventory.title("Inventory")
inventory.protocol("WM_DELETE_WINDOW", inventory.withdraw)
inventory.bind("<Key-Escape>", lambda event: inventory.withdraw())
inventory.resizable(FALSE, TRUE)
inventory.withdraw()

all_buttons["Items"].configure(command=inventory.deiconify)
top.bind("i", lambda event: inventory.deiconify())

inv_toolbar = ttk.Frame(inventory)
inv_toolbar.pack(side=BOTTOM, padx=4, pady=4)

inv_buttons = {}
invkeys = ["Eat", "Wield/wear", "Drop"]
for lbl in invkeys:
	button = ttk.Button(inv_toolbar, text=lbl, underline=0)
	button.pack(side=LEFT)
	inv_buttons[lbl] = button

lbl_coin = ttk.Label(inventory, text="Gold: 0 Silver: 0 Amber: 0")
lbl_coin.pack(side=BOTTOM)

item_display = ttk.Treeview(inventory,
	columns="name amount weight element",
	show="headings", height=15, selectmode="browse")
item_display.heading("name", text="Item")
item_display.heading("amount", text="Amount")
item_display.heading("weight", text="Weight")
item_display.heading("element", text="Element")
item_display.column("name", width=150, minwidth=100)
item_display.column("amount", width=100, minwidth=80, anchor="e")
item_display.column("weight", width=100, minwidth=80, anchor="e")
item_display.column("element", width=100, minwidth=80, anchor="center")
item_display.pack(side=LEFT, fill=BOTH, expand=TRUE)

inv_scroll = ttk.Scrollbar(
	inventory, orient=VERTICAL, command=item_display.yview)
item_display["yscrollcommand"] = inv_scroll.set
inv_scroll.pack(side=RIGHT, fill=Y, expand=TRUE)

charsheet.bind("i", lambda event: inventory.deiconify())
charsheet.bind("<Key-question>", lambda event: helpdlg.deiconify())
inventory.bind("<Key-at>", lambda event: charsheet.deiconify())
inventory.bind("<Key-question>", lambda event: helpdlg.deiconify())
helpdlg.bind("i", lambda event: inventory.deiconify())
helpdlg.bind("<Key-at>", lambda event: charsheet.deiconify())

xp_per_level = 150
xp_attr_cost = 500

size_desc = {4: "a small thing", 6: "smaller than you", 8: "about your size",
	10: "bigger than you", 12: "a large creature", 20: "a huge creature"}

element_glow = {"sun": "red", "plant": "green", "frost": "blue"}
weakness = {"frost": "sun", "sun": "plant", "plant": "frost"}
effect = {"sun": "burning", "plant": "bleeding", "frost": "freezing"}
terrain_heal = {'*': "bleeding", '"': "burning", '&': "freezing"}
terrain_text = {
	'*': "You wash your cuts in the fountain.",
	'"': "You rest in the shadow of the tree.",
	'&': "You warm yourself up at the brazier."
}

terrain = {
	'.': {"name": "the floor", "color": "gray"},
	'#': {"name": "a wall", "color": "firebrick"},
	'|': {"name": "a pillar", "color": "gray50"},
	'*': {"name": "a fountain", "color": "turquoise"},
	'&': {"name": "a brazier", "color": "orangered"},
	'"': {"name": "a tree", "color": "yellowgreen"},
	'=': {"name": "brambles", "color": "darkgreen"},
	'>': {"name": "a way further in", "color": "gray50"},
	'<': {"name": "a way back out", "color": "gray50"},
	'+': {"name": "the fairy king's throne", "color": "gold"}
}

coins = {
	"gold": {
		"char": "$",
		"color": "gold",
		"name": "gold",
		"plural_name": True,
		"max_stack": 100,
		"amount": 10,
		"element": "sun"
	},
	"silver": {
		"char": "$",
		"color": "gray",
		"name": "silver",
		"plural_name": True,
		"max_stack": 100,
		"amount": 10,
		"element": "frost"
	},
	"amber": {
		"char": "$",
		"color": "orangered",
		"name": "amber",
		"plural_name": True,
		"max_stack": 100,
		"amount": 10,
		"element": "plant"
	}
}

foods = {
	"snack": {
		"char": "%",
		"color": "gold",
		"name": "hearty snack",
		"nutrition": 3,
		"weight": 2,
		"uses": 3,
	},
	"fruit": {
		"char": "%",
		"color": crimson,
		"name": "strange fruit",
		"plural_name": True,
		"nutrition": 2,
		"weight": 1,
		"max_stack": 5
	},
	"salad": {
		"char": "%",
		"color": "turquoise",
		"name": "blue salad",
		"plural_name": True,
		"nutrition": 3,
		"weight": 2,
		"max_stack": 3
	},
	"jelly": {
		"char": "%",
		"color": "fuchsia",
		"name": "quivering jelly",
		"plural_name": True,
		"nutrition": 5,
		"weight": 3,
		"max_stack": 2
	},
	"grapes": {
		"char": "%",
		"color": crimson,
		"name": "sour grapes",
		"plural_name": True,
		"nutrition": 2,
		"weight": 4,
		"uses": 3
	},
	"seed": {
		"char": "%",
		"color": crimson,
		"name": "fire seed",
		"plural_name": True,
		"nutrition": 3,
		"weight": 1,
		"max_stack": 5,
		"element": "sun"
	},
	"root": {
		"char": "%",
		"color": "fuchsia",
		"name": "blood root",
		"plural_name": True,
		"nutrition": 4,
		"weight": 2,
		"max_stack": 3,
		"element": "plant"
	},
	"cream": {
		"char": "%",
		"color": "turquoise",
		"name": "ice cream",
		"plural_name": True,
		"nutrition": 6,
		"weight": 3,
		"max_stack": 2,
		"element": "frost"
	},
	"apple": {
		"char": "%",
		"color": "gold",
		"name": "golden apple",
		"nutrition": 8,
		"weight": 4,
		"max_stack": 3
	}
}

weapons = {
	"knife": {
		"char": "(",
		"color": crimson,
		"name": "rusty knife",
		"attack": 3,
		"damage": 4,
		"weight": 1,
		"max_stack": 3
	},
	"shears": {
		"char": "(",
		"color": "gray",
		"name": "garden shears",
		"plural_name": True,
		"attack": 4,
		"damage": 6,
		"weight": 2
	},
	"icicle": {
		"char": ")",
		"color": "turquoise",
		"name": "icicle",
		"attack": 3,
		"damage": 4,
		"weight": 1,
		"max_stack": 3,
		"elemental": True,
		"element": "frost"
	},
	"flame": {
		"char": ")",
		"color": "orangered",
		"name": "lick of flame",
		"attack": 3,
		"damage": 4,
		"max_stack": 7,
		"elemental": True,
		"element": "sun"
	},
	"grass": {
		"char": ")",
		"color": lime,
		"name": "grass blade",
		"attack": 3,
		"damage": 4,
		"nutrition": 1,
		"max_stack": 5,
		"elemental": True,
		"element": "plant"
	},
	"axe": {
		"char": "(",
		"color": "gray",
		"name": "shepherd's axe",
		"attack": 5,
		"damage": 6,
		"weight": 3,
		"max_stack": 2
	},
	"stick": {
		"char": ")",
		"color": lime,
		"name": "thorny stick",
		"attack": 5,
		"damage": 6,
		"weight": 2,
		"max_stack": 2,
		"elemental": True,
		"element": "plant"
	},
	"sunbeam": {
		"char": ")",
		"color": crimson,
		"name": "sunbeam",
		"attack": 5,
		"damage": 6,
		"weight": 0,
		"max_stack": 2,
		"elemental": True,
		"element": "sun"
	},
	"scepter": {
		"char": ")",
		"color": "turquoise",
		"name": "icy scepter",
		"attack": 5,
		"damage": 6,
		"weight": 4,
		"max_stack": 2,
		"elemental": True,
		"element": "frost"
	},
	"staff": {
		"char": "(",
		"color": "firebrick",
		"name": "knotten staff",
		"attack": 7,
		"damage": 8,
		"weight": 7,
		"max_stack": 2
	}
}

# For armor, weight doubles as hitpoints. It absorbs its defense score
# in damage, and takes half of that itself. Falls apart when at zero.
armors = {
	"cloth": {
		"char": "[",
		"color": "gray",
		"name": "faded rags",
		"plural_name": True,
		"defense": 2,
		"weight": 3,
		"max_stack": 3
	},
	"leather": {
		"char": "[",
		"color": crimson,
		"name": "leather apron",
		"defense": 4,
		"weight": 6
	},
	"sheepskin": {
		"char": "[",
		"color": "gray",
		"name": "sheepskin",
		"defense": 4,
		"weight": 8
	},
	"darkness": {
		"char": "[",
		"color": "gray",
		"name": "cloak of darkness",
		"defense": 6,
		"weight": 9
	},
	"mantle": {
		"char": "]",
		"color": lime,
		"name": "leafy mantle",
		"defense": 6,
		"weight": 9,
		"elemental": True,
		"element": "plant"
	},
	"cover": {
		"char": "]",
		"color": "turquoise",
		"name": "snow cover",
		"defense": 6,
		"weight": 9,
		"elemental": True,
		"element": "frost"
	},
	"wreath": {
		"char": "]",
		"color": crimson,
		"name": "wreath of fire",
		"defense": 6,
		"weight": 9,
		"elemental": True,
		"element": "sun"
	},
	"mail": {
		"char": "[",
		"color": "gold",
		"name": "golden mail",
		"defense": 8,
		"weight": 12
	}
}

spells = {
	"burning": {
		"char": "?",
		"color": crimson,
		"name": "burning rhyme",
		"max_stack": 7,
		"uses": 1,
		"target": "weapon",
		"element": "sun"
	},
	"bleeding": {
		"char": "?",
		"color": lime,
		"name": "bleeding rhyme",
		"max_stack": 7,
		"uses": 1,
		"target": "weapon",
		"element": "plant"
	},
	"freezing": {
		"char": "?",
		"color": "turquoise",
		"name": "freezing rhyme",
		"max_stack": 7,
		"uses": 1,
		"target": "weapon",
		"element": "frost"
	},
	"summer": {
		"char": "?",
		"color": crimson,
		"name": "summer rhyme",
		"max_stack": 7,
		"uses": 2,
		"target": "armor",
		"element": "sun"
	},
	"forest": {
		"char": "?",
		"color": lime,
		"name": "forest rhyme",
		"max_stack": 7,
		"uses": 2,
		"target": "armor",
		"element": "plant"
	},
	"winter": {
		"char": "?",
		"color": "turquoise",
		"name": "winter rhyme",
		"max_stack": 7,
		"uses": 2,
		"target": "armor",
		"element": "frost"
	},
	"soothing": {
		"char": "?",
		"color": "gray",
		"name": "soothing rhyme",
		"max_stack": 7,
		"uses": 3,
		"target": "effect"
	},
	"blessing": {
		"char": "?",
		"color": "gold",
		"name": "blessing rhyme",
		"max_stack": 7,
		"uses": 3,
		"target": "fortune"
	}
}

creatures = {
	"human": {
		"char": "@",
		"color": "white",
		"name": "human",
		"dice_size": 8,
		"muscle": 2,
		"stamina": 2,
		"agility": 2,
		"weapon": weapons["knife"],
		"armor": armors["cloth"],
		"spell": spells["blessing"]
	},
	"crawlie": {
		"char": "c",
		"color": "fuchsia",
		"name": "creepy crawlie",
		"dice_size": 4,
		"muscle": 3,
		"stamina": 3,
		"agility": 4,
	},
	"slug": {
		"char": "s",
		"color": crimson,
		"name": "slimy slug",
		"dice_size": 4,
		"muscle": 4,
		"stamina": 4,
		"agility": 3,
	},
	"bug": {
		"char": "b",
		"color": lime,
		"name": "bloated bug",
		"dice_size": 4,
		"muscle": 3,
		"stamina": 4,
		"agility": 3,
	},
	"icesprite": {
		"char": "i",
		"color": "turquoise",
		"name": "ice sprite",
		"dice_size": 6,
		"muscle": 3,
		"stamina": 4,
		"agility": 2,
		"elemental": True,
		"element": "frost",
		"weapon": weapons["icicle"],
		"spell": spells["freezing"],
		"coin": {"gold": 0, "silver": 20, "amber": 0}
	},
	"firesprite": {
		"char": "f",
		"color": "orangered",
		"name": "fire sprite",
		"dice_size": 6,
		"muscle": 2,
		"stamina": 3,
		"agility": 4,
		"elemental": True,
		"element": "sun",
		"weapon": weapons["flame"],
		"spell": spells["burning"],
		"coin": {"gold": 30, "silver": 0, "amber": 0}
	},
	"mandrake": {
		"char": "m",
		"color": lime,
		"name": "mandrake",
		"dice_size": 6,
		"muscle": 3,
		"stamina": 3,
		"agility": 3,
		"elemental": True,
		"element": "plant",
		"weapon": weapons["grass"],
		"spell": spells["bleeding"],
		"coin": {"gold": 0, "silver": 0, "amber": 20}
	},
	"shadow": {
		"char": "S",
		"color": "gray",
		"name": "living shadow",
		"dice_size": 8,
		"muscle": 3,
		"stamina": 3,
		"agility": 3,
		"spell": spells["soothing"],
		"armor": armors["darkness"]
	},
	"wild": {
		"char": "W",
		"color": lime,
		"name": "wild fae",
		"dice_size": 8,
		"muscle": 2,
		"stamina": 2,
		"agility": 3,
		"elemental": True,
		"element": "plant",
		"spell": spells["forest"],
		"armor": armors["mantle"],
		"weapon": weapons["stick"],
		"coin": {"gold": 0, "silver": 0, "amber": 40}
	},
	"maiden": {
		"char": "I",
		"color": "turquoise",
		"name": "ice maiden",
		"dice_size": 8,
		"muscle": 4,
		"stamina": 2,
		"agility": 3,
		"elemental": True,
		"element": "frost",
		"spell": spells["winter"],
		"armor": armors["cover"],
		"weapon": weapons["scepter"],
		"coin": {"gold": 0, "silver": 40, "amber": 0}
	},
	"prince": {
		"char": "L",
		"color": crimson,
		"name": "prince of light",
		"dice_size": 8,
		"muscle": 3,
		"stamina": 3,
		"agility": 3,
		"elemental": True,
		"element": "sun",
		"spell": spells["summer"],
		"armor": armors["wreath"],
		"weapon": weapons["sunbeam"],
		"coin": {"gold": 50, "silver": 0, "amber": 0}
	},
	"wight": {
		"char": "W",
		"color": "gray",
		"name": "wight",
		"dice_size": 8,
		"muscle": 4,
		"stamina": 4,
		"agility": 4,
		"spell": spells["blessing"],
		"undead": True
	},
	"banshee": {
		"char": "B",
		"color": "gray",
		"name": "banshee",
		"dice_size": 8,
		"muscle": 4,
		"stamina": 3,
		"agility": 4,
		"spell": spells["blessing"],
		"undead": True
	},
	"mountman": {
		"char": "M",
		"color": "gray",
		"name": "mountain man",
		"dice_size": 8,
		"muscle": 3,
		"stamina": 3,
		"agility": 3,
		"weapon": weapons["axe"],
		"armor": armors["sheepskin"]
	},
	"hag": {
		"char": "H",
		"color": "gray",
		"name": "hag",
		"dice_size": 8,
		"muscle": 2,
		"stamina": 4,
		"agility": 3,
		"weapon": weapons["stick"],
		"spell": spells["bleeding"]
	},
	"troll": {
		"char": "T",
		"color": lime,
		"name": "forest troll",
		"dice_size": 10,
		"muscle": 3,
		"stamina": 4,
		"agility": 5,
		"elemental": True,
		"element": "plant",
		"spell": spells["forest"],
		"coin": {"gold": 0, "silver": 0, "amber": 70}
	},
	"queen": {
		"char": "Q",
		"color": "turquoise",
		"name": "ice queen",
		"dice_size": 10,
		"muscle": 2,
		"stamina": 2,
		"agility": 3,
		"elemental": True,
		"element": "frost",
		"spell": spells["winter"],
		"armor": armors["cover"],
		"weapon": weapons["scepter"],
		"coin": {"gold": 0, "silver": 70, "amber": 0}
	},
	"hellhound": {
		"char": "H",
		"color": crimson,
		"name": "hellhound",
		"dice_size": 10,
		"muscle": 4,
		"stamina": 4,
		"agility": 4,
		"elemental": True,
		"element": "sun",
		"coin": {"gold": 70, "silver": 0, "amber": 0}
	},
	"warg": {
		"char": "W",
		"color": "firebrick",
		"name": "warg",
		"dice_size": 10,
		"muscle": 4,
		"stamina": 4,
		"agility": 4
	},
	"kelpie": {
		"char": "K",
		"color": "turquoise",
		"name": "kelpie",
		"dice_size": 12,
		"muscle": 4,
		"stamina": 4,
		"agility": 4
	},
	"ogre": {
		"char": "O",
		"color": "gray",
		"name": "ogre",
		"dice_size": 12,
		"muscle": 4,
		"stamina": 4,
		"agility": 4
	},
	"bugbear": {
		"char": "B",
		"color": "firebrick",
		"name": "bugbear",
		"dice_size": 12,
		"muscle": 4,
		"stamina": 4,
		"agility": 4
	},
	"headless": {
		"char": "H",
		"color": "firebrick",
		"name": "headless horseman",
		"dice_size": 12,
		"muscle": 4,
		"stamina": 4,
		"agility": 4,
		"spell": spells["blessing"],
		"undead": True
	},
	"wyvern": {
		"char": "Y",
		"color": "gold",
		"name": "wyvern",
		"dice_size": 12,
		"muscle": 3,
		"stamina": 4,
		"agility": 5
	},
	"icedrake": {
		"char": "D",
		"color": "turquoise",
		"name": "ice drake",
		"dice_size": 20,
		"muscle": 2,
		"stamina": 4,
		"agility": 3,
		"elemental": True,
		"element": "frost",
		"spell": spells["freezing"],
		"coin": {"gold": 0, "silver": 100, "amber": 0}
	},
	"hilldrake": {
		"char": "D",
		"color": lime,
		"name": "hill drake",
		"dice_size": 20,
		"muscle": 3,
		"stamina": 3,
		"agility": 3,
		"elemental": True,
		"element": "plant",
		"spell": spells["bleeding"],
		"coin": {"gold": 0, "silver": 0, "amber": 100}
	},
	"firedrake": {
		"char": "D",
		"color": crimson,
		"name": "fire drake",
		"dice_size": 20,
		"muscle": 3,
		"stamina": 2,
		"agility": 4,
		"elemental": True,
		"element": "sun",
		"spell": spells["burning"],
		"coin": {"gold": 100, "silver": 0, "amber": 0}
	},
	"giant": {
		"char": "G",
		"color": "gray",
		"name": "giant",
		"dice_size": 20,
		"muscle": 4,
		"stamina": 4,
		"agility": 4
	},
	"wyrm": {
		"char": "W",
		"color": "gold",
		"name": "great wyrm",
		"dice_size": 20,
		"muscle": 3,
		"stamina": 4,
		"agility": 5
	}
}

# Unchecked, the level generator adds 3-5 coins per level, 2-3 weapons,
# maybe 6 spells, up to 8 armors and rather too much food. Plan accordingly.
# Note: for good balance, always mix up elemental and mundane creatures.
levels = {
	1: {
		"creatures": ["slug", "bug"],
		"coins": ["gold", "silver", "amber"],
		"weapons": ["shears", None, None],
		"spells": ["blessing", None, None, None, None, None],
		"armors": ["cloth", "cloth", None, None, None, None],
		"foods": ["fruit", "fruit", None, None, None, None],
	},
	2: {
		"creatures": ["crawlie", "slug", "bug", "slug", "bug"],
		"coins": ["gold", "silver", "amber"],
		"weapons": ["shears", None, None],
		"spells": ["summer", "burning", "soothing",
			None, None, None, None, None, None],
		"armors": ["cloth", "cloth", None, None, None],
		"foods": ["fruit", "salad", None, None, None, None]
	},
	3: {
		"creatures": ["slug", "bug", "slug", "bug", "icesprite"],
		"coins": ["gold", None, "amber"],
		"weapons": ["shears", None, None],
		"spells": ["summer", "winter", "freezing",
			"soothing", "blessing", "burning",
			None, None, None, None, None, None,
			None, None, None, None, None, None],
		"armors": ["cloth", "leather", None, None, None, None],
		"foods": ["grapes", "fruit", "salad", None, None, None]
	},
	4: {
		"creatures": ["crawlie", "crawlie", "icesprite", "mandrake"],
		"coins": ["gold", None, None],
		"weapons": ["shears", "axe", None, None, None, None],
		"spells": ["summer", "winter", "forest", "soothing",
			"burning", "freezing", "bleeding", "blessing",
			None, None, None, None, None, None, None, "soothing",
			None, None, None, None, None, None, None, "blessing"],
		"armors": ["cloth", "leather", None, None, None, None],
		"foods": ["grapes", "jelly", "salad", None, None, None]
	},
	5: {
		"creatures": ["crawlie", "crawlie",
			"icesprite", "mandrake", "firesprite"],
		"coins": [None],
		"weapons": ["axe", "axe", None, None, None, None],
		"spells": ["summer", "winter", "forest", "soothing",
			"burning", "freezing", "bleeding", "blessing",
			None, None, None, None, None, None, None, "soothing",
			None, None, None, None, None, None, None, "blessing"],
		"armors": ["leather", "leather", None, None, None, None],
		"foods": ["salad", "jelly", "seed", "root", "cream", None]
	},
	6: {
		"creatures": ["icesprite", "firesprite", "mandrake",
			"shadow", "shadow"],
		"coins": [None],
		"weapons": ["axe", "axe", None, None, None, None],
		"spells": ["summer", "winter", "forest", "soothing",
			"burning", "freezing", "bleeding", "blessing",
			None, None, None, None, None, None, None, "soothing",
			None, None, None, None, None, None, None, "blessing"],
		"armors": ["leather", "sheepskin", None, None, None, None],
		"foods": ["salad", "jelly", "seed", "root", "cream", None]
	},
	7: {
		"creatures": ["icesprite", "firesprite",
			"shadow", "shadow", "wild"],
		"coins": [None],
		"weapons": ["axe", "grass", None, None, None, None],
		"spells": ["soothing", "bleeding",
			"winter", "summer", "forest",
			"burning", "freezing", None, None, None,
			None, None, None, None, None],
		"armors": ["leather", "sheepskin", None, None, None, None],
		"foods": ["salad", "jelly", "seed", "root", "cream", None]
	},
	8: {
		"creatures": ["firesprite", "shadow", "shadow",
			"wild", "maiden"],
		"coins": [None],
		"weapons": ["grass", "icicle", None, None, None, None],
		"spells": ["soothing", "bleeding",
			"winter", "summer", "forest",
			"burning", None, None, None, None,
			None, None, None, None, None],
		"armors": ["leather", "sheepskin", None, None, None, None],
		"foods": ["salad", "jelly", "seed", "root", "cream", None]
	},
	9: {
		"creatures": ["shadow", "shadow", "wight", "banshee", None],
		"coins": ["gold", "silver", "amber"],
		"weapons": ["grass", "icicle", "flame", None, None, None],
		"spells": ["soothing", "bleeding",
			"winter", "summer", "forest",
			None, None, None, None, None,
			None, None, None, None, None],
		"armors": ["leather", "sheepskin", None, None, None, None],
		"foods": ["seed", "root", "cream", None, None, None]
	},
	10: {
		"creatures": ["banshee", "wild", "maiden", "prince", None],
		"coins": [None],
		"weapons": ["axe", None, None],
		"spells": ["bleeding", "freezing", "burning", "soothing",
			"winter", "summer", "forest", "blessing",
			None, None, None, None, None, None, None, "soothing",
			None, None, None, None, None, None, None, "blessing"],
		"armors": ["sheepskin", "darkness", None, None, None, None],
		"foods": ["fruit", "salad", "jelly", None, None, None]
	},
	11: {
		"creatures": ["wild", "maiden", "prince",
			"mountman", None, "mountman", None],
		"coins": [None],
		"weapons": ["axe", None, None],
		"spells": ["bleeding", "freezing", "burning", "soothing",
			"winter", "summer", "forest", "blessing",
			None, None, None, None, None, None, None, "soothing",
			None, None, None, None, None, None, None, "blessing"],
		"armors": ["sheepskin", "darkness", None, None, None, None],
		"foods": ["salad", "jelly", "seed", "root", "cream", None]
	},
	12: {
		"creatures": ["maiden", "prince", "hag",
			"mountman", None, "mountman", None],
		"coins": [None, None, "amber"],
		"weapons": ["axe", None, None],
		"spells": ["bleeding", "freezing", "burning",
			"soothing", "blessing", "summer", "winter",
			None, None, None, None, None, None, None,
			None, None, None, None, None, None, None],
		"armors": ["sheepskin", "mantle", None, None, None, None],
		"foods": ["salad", "jelly", "seed", "root", "cream", None]
	},
	13: {
		"creatures": ["queen", "prince", "hag",
			"mountman", None, "mountman", None],
		"coins": [None, None, "amber"],
		"weapons": ["axe", None, None],
		"spells": ["bleeding", "freezing", "burning",
			"soothing", "blessing", "summer", "winter",
			None, None, None, None, None, None, None,
			None, None, None, None, None, None, None],
		"armors": ["sheepskin", "mantle", None, None, None, None],
		"foods": ["salad", "jelly", "seed", "root", "cream", None]
	},
	14: {
		"creatures": ["queen", "hellhound", "hag",
			"mountman", None, "mountman", None],
		"coins": [None, None, "amber"],
		"weapons": ["sunbeam", None, None],
		"spells": ["bleeding", "freezing", "winter",
			"soothing", "blessing", None,
			None, None, None,
			None, None, None],
		"armors": ["wreath", "mantle", None, None, None, None],
		"foods": ["salad", "jelly", "seed", "root", "cream", None]
	},
	15: {
		"creatures": ["queen", "hellhound", "hag",
			"warg", None, "warg", None],
		"coins": [None, None, "amber"],
		"weapons": ["sunbeam", None, None],
		"spells": ["bleeding", "freezing", "winter",
			"soothing", "blessing", None,
			None, None, None,
			None, None, None],
		"armors": ["wreath", "mantle", None, None, None, None],
		"foods": ["apple", "jelly", "seed", "root", "cream", None]
	},
	16: {
		"creatures": ["queen", "hellhound", "troll",
			"warg", None, "warg", None],
		"coins": [None, None, None],
		"weapons": ["stick", "sunbeam", "staff",
			None, None, None, None, None, None],
		"spells": ["winter", "freezing", None,
			"soothing", "blessing", None,
			None, None, None,
			None, None, None],
		"armors": ["wreath", "mantle", None, None, None, None],
		"foods": ["apple", "jelly", "seed", "root", "cream", None]
	},
	17: {
		"creatures": ["ogre", "hellhound", "troll",
			"warg", None, "warg", None],
		"coins": [None, "silver", None],
		"weapons": ["stick", "sunbeam", "scepter",
			None, None, None, None, None, None],
		"spells": ["soothing", "blessing", None, None, None, None],
		"armors": ["wreath", "mantle", "cover",
			None, None, None, None, None, None],
		"foods": ["apple", "jelly", "cream", None, None, None]
	},
	18: {
		"creatures": ["bugbear", "bugbear", "bugbear",
			"headless", "headless", None],
		"coins": ["gold", "silver", "amber"],
		"weapons": ["staff", None, None],
		"spells": ["bleeding", "freezing", "burning",
			"soothing", None, None,
			None, None, None],
		"armors": ["mail", None, None],
		"foods": ["seed", "root", "cream", None, None, None]
	},
	19: {
		"creatures": ["ogre", "bugbear", "headless", "kelpie", None],
		"coins": ["gold", "silver", "amber"],
		"weapons": ["stick", "sunbeam", "scepter",
			None, None, None, None, None, None],
		"spells": ["soothing", "blessing", None, None, None, None],
		"armors": ["mail", None, None],
		"foods": ["apple", "jelly", None, None]
	},
	20: {
		"creatures": ["ogre", "bugbear", "wyvern", "kelpie", None],
		"coins": ["gold", "silver", "amber"],
		"weapons": ["stick", "sunbeam", "scepter",
			None, None, None, None, None, None],
		"spells": ["soothing", "blessing", None, None, None, None],
		"armors": ["mail", None, None],
		"foods": ["apple", "jelly", None, None]
	},
	21: {
		"creatures": ["icedrake", "bugbear", "wyvern", "kelpie", None],
		"coins": ["gold", None, "amber"],
		"weapons": ["stick", "sunbeam", "staff",
			None, None, None, None, None, None],
		"spells": ["soothing", "blessing", "winter",
			"forest", None, None, None, None, None],
		"armors": ["mail", "wreath", None, None, None, None],
		"foods": ["apple", "jelly", "cream", None, None, None]
	},
	22: {
		"creatures": ["icedrake", "hilldrake",
			"wyvern", "kelpie", None],
		"coins": ["gold", None, None],
		"weapons": ["staff", "sunbeam", "staff",
			None, None, None, None, None, None],
		"spells": ["soothing", "blessing", None,
			"forest", None, None, None, None, None],
		"armors": ["mail", "wreath", "cover",
			None, None, None, None, None, None],
		"foods": ["apple", "jelly", "cream", "root", None, None]
	},
	23: {
		"creatures": ["icedrake", "hilldrake",
			"wyvern", "firedrake", None],
		"coins": [None],
		"weapons": ["staff", None, None],
		"spells": ["soothing", "blessing", None, None, None, None],
		"armors": ["wreath", "mantle", "cover",
			None, None, None, None, None, None],
		"foods": ["apple", "jelly", "cream", "root", "seed", None]
	},
	24: {
		"creatures": ["icedrake", "hilldrake",
			"giant", "firedrake", None],
		"coins": [None],
		"weapons": ["staff", None, None],
		"spells": ["soothing", "blessing", None, None, None, None],
		"armors": ["wreath", "mantle", "cover",
			None, None, None, None, None, None],
		"foods": ["apple", "jelly", "cream", "root", "seed", None]
	},
	25: {
		"creatures": ["wyrm", "giant", None],
		"coins": ["gold", "silver", "amber"],
		"weapons": ["staff", None, None],
		"spells": ["soothing", "blessing", None, None, None, None],
		"armors": ["mail", None, None],
		"foods": ["apple", "jelly", None, None]
	}
}

defeat_msg = """As you endure one blow too many, your vision fades to black and your knees buckle.

Time passes.

You wake up on the bank of a stream that flows merrily uphill, crossed by a small rainbow that looks solid enough to walk on. Trees look down on you with their hollow eyes, while shadows without a source chase each other through the grass.

As you watch, the shadows circle closer and closer, gradually taking shape.

You are now a plaything of the fairies."""

victory_msg = """No-one stops you upon coming near the throne. Walking around it soon reveals the reason: it's empty. You wait for long moments, but nothing happens. A mischievous thought comes to you, and you sit down. Only then do you notice the crown resting on a small table nearby... and covered in dust.

The Fairy King has been absent for a long time. You will not get your wish after all.

But look at you! Laden with riches, and powerful enough to best dragons. No longer the person you came in as. The largest reward anyone can hope for.

You get up to begin the long journey home."""

def next_coords(x, y, direction):
	if direction == 'n':
		return x, y - 1
	elif direction == 's':
		return x, y + 1
	elif direction == 'e':
		return x + 1, y
	elif direction ==  'w':
		return x - 1, y
	else:
		raise ValueError("Invalid compass direction " + str(direction))

def roll_dice(count, sides):
	roll = 0
	for i in range(count):
		roll += random.randint(1, sides)
	return roll

def blessed_roll(count, sides):
	return max(roll_dice(count, sides), roll_dice(count, sides))

def cursed_roll(count, sides):
	return min(roll_dice(count, sides), roll_dice(count, sides))

class Dungeon:
	def __init__(self, generator):
		self.player = Creature(creatures["human"])
		self.player.pronoun = "them"
		self.player.proper_name = True
		self.player.acquire(Item(foods["snack"]))
		self.level1 = DungeonLevel(1)
		self.level1.populate(generator)
		
		x, y = self.level1.stairs_down
		self.player.move_to(x, y, self.level1)

	def update(self):
		p = self.player
		del p.in_rooms[:]
		for i in p.level.rooms:
			if i.contains(p.x, p.y):
				p.in_rooms.append(i)
				i.mark_seen(p.level.level_map)
		for i in p.level.content[:]:
			if not isinstance(i, Creature):
				continue
			elif i.wounds < i.health():
				continue
			i.drop_things()
			p.level.content.remove(i)
			if i == p:
				p.log("You are defeated!! (press any key)")
			else:
				name = i.the_name().capitalize()
				p.log("%s flees, defeated!" % (name,))

class DungeonLevel:
	width = 40
	height = 25
	
	def __init__(self, depth):
		self.depth = depth

		self.level_map = [[None] * self.width
			for i in range(self.height)]
		self.prev_level = None
		self.next_level = None
		
		self.rooms = []
		self.doorways = []
		self.content = []
	
	def __str__(self):
		return "\n".join(
			["".join([str(j) for j in i])
				for i in self.level_map])
	
	def is_on_grid(self, x, y):
		return x >= 0 and y >= 0 and x < self.width and y < self.height

	def any_items_at(self, x, y):
		for i in self.content:
			if isinstance(i, Item) and i.x == x and i.y == y:
				return True
		return False
	
	def items_at(self, x, y):
		return (i for i in self.content
			if isinstance(i, Item) and i.x == x and i.y == y)
	
	def creature_at(self, x, y):
		for i in self.content:
			if isinstance(i, Creature) and i.x == x and i.y == y:
				return i
		return None
	
	def dig_next(self):
		if self.next_level != None:
			return
		
		self.next_level = DungeonLevel(self.depth + 1)
		self.next_level.prev_level = self
		self.next_level.populate(self.generator)
	
	def populate(self, generator):
		self.generator = generator
		generator.populate(self)

class MapTile:
	def __init__(self, char = "."):
		self.char = char
		self.seen = False
		self.icon_id = 0
	
	def __str__(self):
		return self.char

class Room:
	def __init__(self, x1, y1, x2, y2):
		self.x1 = x1
		self.y1 = y1
		self.x2 = x2
		self.y2 = y2
	
	def __str__(self):
		return ("Room %d,%d - %d,%d"
			% (self.x1, self.y1, self.x2, self.y2))
	
	def contains(self, x, y):
		return (x >= self.x1 and y >= self.y1
			and x <= self.x2 and y <= self.y2)

	def mark_seen(self, level_map):
		for y in range(self.y1, self.y2 + 1):
			for x in range(self.x1, self.x2 + 1):
				level_map[y][x].seen = True

class Thing:
	def __init__(self):
		self.level = None
		self.x = -1
		self.y = -1
		
		self.name = "thing"
		self.pronoun = "it"
		self.proper_name = False
		self.plural_name = False
		self.elemental = False
		self.element = None
		
		self.icon_id = 0
	
	def move_to(self, x, y, parent = None):
		l = self.level
		
		if hasattr(l, "content") and self in l.content:
			l.content.remove(self)
		if hasattr(parent, "content"):
			parent.content.append(self)
			self.level = parent
		
		if not isinstance(parent, DungeonLevel): return

		if self.level.is_on_grid(x, y):
			self.x = x
			self.y = y
		else:
			raise ValueError("Can't move outside the level map.")

	def the_name(self):
		if self.proper_name:
			return self.name
		else:
			return "the " + self.name
	
	def a_name(self):
		if self.plural_name:
			return "some " + self.name
		elif self.name[0] in ('a', 'e', 'i', 'o', 'u'):
			return "an " + self.name
		elif self.proper_name:
			return self.name
		else:
			return "a " + self.name

	def __str__(self):
		return self.name

class Item(Thing):
	def __init__(self, template):
		Thing.__init__(self)
		self.char = "?"
		
		self.nutrition = 0
		self.attack = 0
		self.defense = 0
		self.damage = 0
		self.uses = 1
		self.target = None
		
		self.weight = 0
		self.max_stack = 1
		self.amount = 1
		
		for attr in template.keys():
			self.__dict__[attr] = template[attr]
		self.template = template
	
	def space_left(self):
		return self.max_stack - self.amount
	
	def can_stack_with(self, other):
		if self.name != other.name:
			return False
		elif self.weight != other.weight:
			return False
		elif self.uses != other.uses:
			return False
		elif self.element != other.element:
			return False
		else:
			return True
	
	@staticmethod
	def from_coin(kind, amount):
		assert kind in coins, "Bad kind of coin: " + kind
		coin = Item(coins[kind])
		coin.amount = amount
		return coin
	
class Creature(Thing):
	def __init__(self, template):
		Thing.__init__(self)
		self.log = print if not quiet else lambda msg: None

		self.char = "@"
		self.dice_size = 8

		self.muscle = 3
		self.agility = 3
		self.stamina = 3
		
		self.weapon = None
		self.armor = None
		self.spell = None
		
		self.can_fly = False
		self.undead = False
		
		self.coin = {"gold": 0, "silver": 0, "amber": 0}
		
		for attr in template.keys():
			if attr == "coin":
				for i in template["coin"]:
					self.coin[i] = template["coin"][i]
			elif attr == "weapon":
				self.weapon = Item(template["weapon"])
			elif attr == "armor":
				self.armor = Item(template["armor"])
			elif attr == "spell":
				self.spell = Item(template["spell"])
			else:
				self.__dict__[attr] = template[attr]
		self.template = template
		
		self.wounds = 0
		self.total_xp = 0
		self.unused_xp = 0
		self.blessings = 0
		self.curses = 0
		
		self.content = []
		self.effects = {}
		self.in_rooms = []
		
		self.moves = 0
		self.victorious = False
	
	def health(self):
		return self.stamina * self.dice_size
	
	def can_see(self, x, y):
		for i in self.in_rooms:
			if i.x1 <= x <= i.x2 and i.y1 <= y <= i.y2:
				return True
		return False
	
	def look_at(self, x, y):
		if not self.can_see(x, y):
			self.log("You can't see over there right now.")
			return

		msg = "You see: "

		mob = self.level.creature_at(x, y)
		if mob != None and mob != self:
			msg += "%s, %s" % (
				mob.a_name(), size_desc[mob.dice_size])
			if mob.element != None:
				msg += " and glowing %s" % (
					element_glow[mob.element])
			elif mob.undead:
				msg += " and dead alive"
			self.log(msg + ".")
			return

		items = list(self.level.items_at(x, y))
		if len(items) > 0:
			names = (i.a_name() for i in items)
			msg += "; ".join(names)
			self.log(msg + ".")
		elif mob == self:
			self.log(msg + "yourself.")
		else:
			tile = self.level.level_map[y][x].char
			msg += terrain[tile]["name"]
			self.log(msg + ".")
	
	def can_enter(self, x, y):
		tile = self.level.level_map[y][x].char
		if tile in ('.', '>', '<', '+'):
			return True
		elif tile == '=':
			return self.can_fly
		else:
			return False

	def walk(self, direction):
		assert self.level != None, "Creature is out of play."
		
		new_x, new_y = next_coords(self.x, self.y, direction)

		if not self.level.is_on_grid(new_x, new_y):
			self.log("An invisible barrier stops you.")
			return False
		elif self.level.creature_at(new_x, new_y) != None:
			creature = self.level.creature_at(new_x, new_y)
			if self.is_enemy(creature):
				self.fight(creature)
			else:
				self.log("You bump into " + creature.name)
			return True
		elif not self.can_enter(new_x, new_y):
			tile = self.level.level_map[new_y][new_x]
			if tile.char in terrain_heal:
				msg1 = terrain_text[tile.char]
				fx = terrain_heal[tile.char]
				if fx in self.effects:
					del self.effects[fx]
					msg2 = "You are no longer %s." % (fx,)
				else:
					msg2 = "It feels good."
				self.log("%s %s" % (msg1, msg2))
			else:
				name = terrain[tile.char]["name"]
				self.log("The way is barred by %s." % (name,))
			return False
		else:
			self.x = new_x
			self.y = new_y
			self.moves += 1
			return True
	
	def is_enemy(self, opponent):
		return self.name != opponent.name
	
	def fight(self, they):
		if self.blessings > 0:
			my_roll = blessed_roll(self.agility, self.dice_size)
			self.blessings -= 1
		elif self.curses > 0:
			my_roll = cursed_roll(self.agility, self.dice_size)
			self.curses -= 1
		else:
			my_roll = roll_dice(self.agility, self.dice_size)
		
		if self.weapon != None:
			my_roll += self.weapon.attack
		
		their_roll = roll_dice(they.agility, they.dice_size)
		if they.weapon != None:
			their_roll += they.weapon.attack
		
		if self.dice_size < they.dice_size:
			my_roll += (they.dice_size - self.dice_size) // 2
		elif self.dice_size > they.dice_size:
			their_roll += (self.dice_size - they.dice_size) // 2
		
		if __debug__:
			print("Combat roll: %d vs. %d"
				% (my_roll, their_roll))
		
		attack = random.choice(["lunges", "swings", "strikes"])
		msg1 = "%s %s at %s" % (
			self.the_name().capitalize(),
			attack, they.the_name())
		if my_roll > their_roll:
			dmg = self.injure(they, my_roll - their_roll)
			msg2 = "and lands a hit! %d damage!" % (dmg,)
		elif their_roll > my_roll:
			they.injure(self, their_roll - my_roll)
			msg2 = "but the latter turns the tables!"
		else:
			msg2 = "but the latter defends well."
		msg = "%s, %s" % (msg1, msg2)
		self.log(msg)
		they.log(msg)

	def injure(self, they, margin):
		damage = margin + self.muscle_bonus()
		def_el = they.defense_element()

		if self.weapon != None:
			att_el = self.weapon.element
			if def_el == None or att_el == None:
				dmg_die = self.weapon.damage
			elif weakness[def_el] == att_el:
				dmg_die = self.weapon.damage * 2
			elif weakness[att_el] == def_el:
				dmg_die = self.weapon.damage // 2
			else:
				dmg_die = self.weapon.damage
			if __debug__:
				print("Rolling 1d%d for damage." % (dmg_die,))
			damage += roll_dice(1, dmg_die)
		else:
			att_el = self.element
			if def_el == None or att_el == None:
				pass
			elif weakness[def_el] == att_el:
				damage *= 2
			elif weakness[att_el] == def_el:
				damage //= 2

		if they.armor != None and damage >= they.armor.defense:
			if __debug__:
				print("Armor: %d" % (they.armor.defense,))
			damage -= they.armor.defense
			they.armor.weight -= they.armor.defense // 2
			if they.armor.weight <= 0:
				they.armor = None
				they.log("Your armor falls apart.")
		
		if __debug__:
			print("Damage: %d" % (damage,))
		
		if damage >= 0:
			damage += they.apply_effect(att_el)
			they.wounds += damage
			halfdie = they.dice_size // 2
			if self.undead:
				they.curses += damage // halfdie
			else:
				they.blessings += damage // halfdie
			return damage
		else:
			return 0
	
	def apply_effect(self, element):
		if element not in effect:
			return 0
		elif self.element == element:
			return 0
		elif self.undead:
			return 0
		elif random.randint(0, 1) == 0:
			return 0
		
		fx = effect[element]
		
		if fx not in self.effects:
			self.effects[fx] = 0
			self.log("You are %s." % (fx,))
		
		self.effects[fx] += 1
		return self.effects[fx]
	
	def muscle_bonus(self):
		return (self.muscle - 3) * self.dice_size // 2
	
	def attack_bonus(self):
		return self.weapon.attack if self.weapon != None else 0
	
	def damage_bonus(self):
		return self.weapon.damage if self.weapon != None else 0
	
	def defense_bonus(self):
		return self.armor.defense if self.armor != None else 0
	
	def defense_element(self):
		if self.armor != None:
			return self.armor.element
		else:
			return self.element
	
	def ascend(self):
		tile = self.level.level_map[self.y][self.x].char
		if tile == '>':
			if self.level.next_level == None:
				self.level.dig_next()
				self.total_xp += xp_per_level
				self.unused_xp += xp_per_level
				if self.wounds >= self.dice_size // 2:
					self.wounds -= self.dice_size // 2
				if self.unused_xp >= xp_attr_cost:
					self.log("You feel more experienced.")
			dest = self.level.next_level
			new_x, new_y = dest.stairs_down
			self.move_to(new_x, new_y, dest)
			self.moves += 1
		elif tile == '+':
			self.victorious = True
			self.log("You are victorious! (press any key)")
		else:
			self.log("There's no way up from here.")
		
	def descend(self):
		tile = self.level.level_map[self.y][self.x].char
		if tile != '<':
			self.log("There's no way down from here.")
		elif self.level.depth > 1:
			dest = self.level.prev_level
			new_x, new_y = dest.stairs_up
			self.move_to(new_x, new_y, dest)
			self.moves += 1
		else:
			self.log("You can't leave without your prize.")
	
	def get_item_here(self):
		item = next(self.level.items_at(self.x, self.y), None)
		if item == None:
			self.log("There's nothing portable here.")
			return
		elif item.weight == 0:
			pass
		elif self.carried_weight() >= self.carrying_capacity():
			self.log("You can't carry any more.")
			return
		
		item.move_to(-1, -1)
		self.acquire(item)
		
		if item.name in self.coin:
			self.log("You pick up %d %s." % (
				item.amount, item.name))
		elif item.amount > 1:
			self.log("You pick up %s times %d." % (
				item.a_name(), item.amount))
		else:
			self.log("You pick up %s." % (item.the_name(),))

	def carried_weight(self):
		return sum(i.weight * i.amount for i in self.content)

	def carrying_capacity(self):
		return self.muscle * self.dice_size * 3

	def acquire(self, item):
		if item.name in self.coin:
			xp_gain = item.amount - self.coin[item.name] // 10
			self.total_xp += xp_gain
			self.unused_xp += xp_gain
			self.coin[item.name] += item.amount
			if self.unused_xp >= xp_attr_cost:
				self.log("You feel more experienced.")
		else:
			for i in self.content:
				if not item.can_stack_with(i):
					continue
				elif item.amount > i.space_left():
					continue
				else:
					i.amount += item.amount
					return
			self.content.append(item)
			item.parent = self

	def eat(self, slot):
		item = self.content[slot]
		
		if item.nutrition > 0:
			if self.wounds < item.nutrition:
				self.log("You're not that hungry. Yet.")
				return
			item.uses -= 1
			if item.uses <= 0:
				self.drop(slot, 1)
			self.wounds -= item.nutrition
			self.apply_effect(item.element)
			assert self.wounds >= 0
			self.log("You feel a little better after eating "
				+ item.the_name() + ".")
		elif item.plural_name:
			self.log(item.the_name().capitalize()
				+ " don't seem edible.")
		else:
			self.log(item.the_name().capitalize()
				+ " doesn't seem edible.")

	def equip(self, slot):
		# slot stands for inventory slot here, not equipment slot
		item = self.content[slot]
		
		if item.attack > 0:
			old_weapon = self.weapon
			self.weapon = self.drop(slot, 1)
			if old_weapon != None:
				self.acquire(old_weapon)
				old_weapon.parent = self
			self.log("You wield " + item.a_name() + ".")
		elif item.defense > 0:
			old_armor = self.armor
			self.armor = self.drop(slot, 1)
			if old_armor != None:
				self.acquire(old_armor)
				old_armor.parent = self
			self.log("You wear " + item.a_name() + ".")
		elif item.uses > 0 and item.target != None:
			old_spell = self.spell
			self.spell = self.drop(slot, 1)
			if old_spell != None:
				self.acquire(old_spell)
				old_spell.parent = self
			self.log("You memorize " + item.a_name() + ".")
		elif item.plural_name:
			self.log(item.the_name().capitalize()
				+ " don't look like a piece of gear.")
		else:
			self.log(item.the_name().capitalize()
				+ " doesn't look like a piece of gear.")

	def drop(self, slot, amount):
		try:
			item = self.content[slot]
		except IndexError:
			print("Tried to drop item not in inventory.",
				file=sys.stderr)
			return None
		if amount > item.amount:
			self.log("But you only have %d!" % (item.amount))
			return None
		else:
			item.amount -= amount
			if item.amount < 1:
				del self.content[slot]
			dropped = Item(item.template)
			dropped.amount = amount
			return dropped

	def drop_things(self):
		# Drop light items first so players can always take them.
		for i in self.coin:
			if self.coin[i] > 0:
				coin = Item.from_coin(i, self.coin[i])
				coin.move_to(self.x, self.y, self.level)
				self.coin[i] = 0
		if self.spell != None and random.randint(0, 1) == 0:
			self.spell.move_to(self.x, self.y, self.level)
			self.spell = None
		if self.armor != None and random.randint(0, 1) == 0:
			self.armor.move_to(self.x, self.y, self.level)
			self.armor = None
		if self.weapon != None and random.randint(0, 1) == 0:
			self.weapon.move_to(self.x, self.y, self.level)
			self.weapon = None

	def cast_spell(self):
		if self.spell == None:
			self.log("But you have no spell prepared.")
		elif self.spell.target == "weapon":
			if self.enchant_slot("weapon", self.spell.element):
				self.update_spell()
		elif self.spell.target == "armor":
			if self.enchant_slot("armor", self.spell.element):
				self.update_spell()
		elif self.spell.target == "effect":
			fx = list(self.effects.keys())
			if len(fx) > 0:
				fx = fx[0]
				del self.effects[fx]
				self.log("You are no longer %s." % (fx,))
			elif self.wounds > 0:
				self.wounds -= 1
				self.log("You feel a little better.")
			else:
				self.log("Nothing seems to happen.")
			self.update_spell()
		elif self.spell.target == "fortune":
			if self.curses > 0:
				self.log("You get a weight off your chest.")
				self.curses = 0
			else:
				self.log("You're feeling truly blessed.")
				self.blessings += 2
			self.update_spell()
		else:
			raise ValueError(
				"Bad spell target: "
					+ str(self.spell.target))

	def update_spell(self):
		self.spell.uses -= 1
		if self.spell.uses <= 0:
			self.log("Your %s scatters into the wind."
				% (self.spell.name))
			self.spell = None

	def enchant_slot(self, slot, element):
		item = self.__dict__[slot]
		if item == None:
			self.log("The spell dies on your lips."
				+ " You have no %s." % (slot,))
			return False
		elif item.elemental:
			if item.element == element:
				if item.plural_name:
					self.log("Your %s glow brighter."
						% (item.name,))
				else:
					self.log("Your %s glows brighter."
						% (item.name,))
			else:
				if item.plural_name:
					self.log(("Your %s sizzle and"
						+ " goes up in smoke.")
							% (item.name,))
				else:
					self.log(("Your %s sizzles and"
						+ " goes up in smoke.")
							% (item.name,))
				self.__dict__[slot] = None
			return True
		elif item.element == element:
			self.log("The glow of your %s fades away."
				% (item.name,))
			item.element = None
			return True
		else:
			if item.plural_name:
				self.log("Your %s acquire a %s glow."
					% (item.name, element_glow[element]))
			else:
				self.log("Your %s acquires a %s glow."
					% (item.name, element_glow[element]))
			item.element = element
			return True

class PalaceFloorGenerator:
	def __init__(self, rng):
		self.rng = rng
	
	def populate(self, level):
		self.level = level
		
		for x in range(level.width):
			for y in range(level.height):
				level.level_map[y][x] = MapTile()
		for x in range (level.width):
			level.level_map[0][x] = MapTile('#')
			level.level_map[-1][x] = MapTile('#')
		for y in range(level.height):
			level.level_map[y][0] = MapTile('#')
			level.level_map[y][-1] = MapTile('#')

		self.subdivide_width(1, 1, level.width - 2, level.height - 2)
		
		stairs = [(2, 2), (level.width - 3, 2),
			(level.width - 3, level.height - 3),
			(2, level.height - 3)]
		i = self.rng.randint(0, len(stairs) - 1)
		level.stairs_down = stairs[i]
		level.stairs_up = stairs[(i + 2) % 4]
		x1, y1 = level.stairs_down
		x2, y2 = level.stairs_up
		level.level_map[y1][x1] = MapTile('<')
		if level.depth < 25:
			level.level_map[y2][x2] = MapTile('>')
		else:
			level.level_map[y2][x2] = MapTile('+')
		if __debug__:
			print("Level %d has %d rooms and %d doorways."
				% (level.depth, len(level.rooms), len(level.doorways)))
		
		if level.depth not in levels:
			return

		depth = level.depth
		for i in range(len(level.doorways)):
			x, y = level.doorways[i]
			if i % 11 == 0:
				kind = self.pick_thing("coins")
				if kind != None:
					thing = Item(coins[kind])
					thing.amount = (
						depth - depth // 2) * 10
					thing.move_to(x, y, level)
			elif i % 13 == 0:
				kind = self.pick_thing("weapons")
				if kind != None:
					thing = Item(weapons[kind])
					thing.move_to(x, y, level)
			elif i % 7 == 0:
				kind = self.pick_thing("spells")
				if kind != None:
					thing = Item(spells[kind])
					thing.move_to(x, y, level)
			elif i % 5 == 0:
				kind = self.pick_thing("armors")
				if kind != None:
					thing = Item(armors[kind])
					thing.move_to(x, y, level)
			elif i % 3 == 0:
				kind = self.pick_thing("foods")
				if kind != None:
					thing = Item(foods[kind])
					thing.move_to(x, y, level)
			elif i % 2 == 0 or self.rng.randint(1, 3) <= 2:
				kind = self.pick_thing("creatures")
				if kind != None:
					thing = Creature(creatures[kind])
					thing.move_to(x, y, level)
	
	def subdivide(self, x1, y1, x2, y2):
		w = x2 - x1 + 1
		h = y2 - y1 + 1

		if w == h and w <= 11:
			self.furnish_room(x1, y1, x2, y2)
			return False
		elif w >= h and w >= 7:
			if (self.level.depth % 3) == 0 and (h == 3 or h == 5):
				self.furnish_room(x1, y1, x2, y2)
				return False
			else:
				return self.subdivide_width(x1, y1, x2, y2)
		elif h >= 7:
			return self.subdivide_height(x1, y1, x2, y2)
		else:
			self.furnish_room(x1, y1, x2, y2)
			return False
	
	def subdivide_width(self, x1, y1, x2, y2):
		w = x2 - x1 + 1
		h = y2 - y1 + 1

		x = x1 + self.rng.randint(3, w - 4)

		for y in range(y1, y2 + 1):
			self.level.level_map[y][x] = MapTile('#')

		self.subdivide(x1, y1, x - 1, y2)
		self.subdivide(x + 1, y1, x2, y2)
		
		doory = y1 + 1 + self.rng.randint(0, h - 2)

		self.level.level_map[doory][x] = MapTile('.')
		# Account for walls placed deeper into recursion.
		self.level.level_map[doory][x - 1] = MapTile('.')
		self.level.level_map[doory][x + 1] = MapTile('.')
		
		self.level.doorways.append((x, doory))
		return True

	def subdivide_height(self, x1, y1, x2, y2):
		w = x2 - x1 + 1
		h = y2 - y1 + 1

		y = y1 + self.rng.randint(3, h - 4)

		for x in range(x1, x2 + 1):
			self.level.level_map[y][x] = MapTile('#')

		self.subdivide(x1, y1, x2, y - 1)
		self.subdivide(x1, y + 1, x2, y2)
		
		doorx = x1 + 1 + self.rng.randint(0, w - 2)

		self.level.level_map[y][doorx] = MapTile('.')
		# Account for walls placed deeper into recursion.
		self.level.level_map[y - 1][doorx] = MapTile('.')
		self.level.level_map[y + 1][doorx] = MapTile('.')
			
		self.level.doorways.append((doorx, y))
		return True
	
	def furnish_room(self, x1, y1, x2, y2):
		w = x2 - x1 + 1
		h = y2 - y1 + 1
		lmap = self.level.level_map
		
		self.level.rooms.append(Room(x1 - 1, y1 - 1, x2 + 1, y2 + 1))
		
		if w == 3 and h == 3:
			if self.level.depth % 9 == 0:
				lmap[y1 + 1][x1 + 1] = MapTile('#')
			else:
				tile = self.rng.choice(['*', '&', '"'])
				lmap[y1 + 1][x1 + 1] = MapTile(tile)
		elif w == 3:
			if h % 2 == 1:
				for y in range(y1 + 1, y2, 2):
					lmap[y][x1 + 1] = MapTile('|')
			elif (self.level.depth % 9) == 0 and h == 4:
				for y in range(y1 + 1, y2):
					lmap[y][x1 + 1] = MapTile('#')
			else:
				for y in range(y1 + 1, y2):
					lmap[y][x1 + 1] = MapTile('=')
		elif h == 3:
			if w % 2 == 1:
				for x in range(x1 + 1, x2, 2):
					lmap[y1 + 1][x] = MapTile('|')
			elif (self.level.depth % 9) == 0 and w == 4:
				for x in range(x1 + 1, x2):
					lmap[y1 + 1][x] = MapTile('#')
			else:
				for x in range(x1 + 1, x2):
					lmap[y1 + 1][x] = MapTile('=')
		elif w == 4 and h == 4:
			lmap[y1 + 1][x1 + 1] = MapTile('=')
			lmap[y1 + 1][x2 - 1] = MapTile('=')
			lmap[y2 - 1][x1 + 1] = MapTile('=')
			lmap[y2 - 1][x2 - 1] = MapTile('=')
		elif w == 5 and h == 5:
			if self.level.depth % 9 == 0:
				tile = self.rng.choice(['|', '#'])
			else:
				tile = self.rng.choice(
					['|', '#', '*', '&', '"'])
			lmap[y1 + 1][x1 + 1] = MapTile(tile)
			lmap[y1 + 1][x2 - 1] = MapTile(tile)
			lmap[y2 - 1][x1 + 1] = MapTile(tile)
			lmap[y2 - 1][x2 - 1] = MapTile(tile)
		elif w == 5:
			if h % 2 == 1:
				for y in range(y1 + 1, y2, 2):
					lmap[y][x1 + 1] = MapTile('|')
					lmap[y][x2 - 1] = MapTile('|')
			else:
				for y in range(y1 + 1, y2):
					lmap[y][x1 + 1] = MapTile('=')
					lmap[y][x2 - 1] = MapTile('=')
		elif h == 5:
			if w % 2 == 1:
				for x in range(x1 + 1, x2, 2):
					lmap[y1 + 1][x] = MapTile('|')
					lmap[y2 - 1][x] = MapTile('|')
			else:
				for x in range(x1 + 1, x2):
					lmap[y1 + 1][x] = MapTile('=')
					lmap[y2 - 1][x] = MapTile('=')
		elif w == 7 and h == 7:
			for y in range(y1 + 1, y2, 2):
				for x in range(x1 + 1, x2, 2):
					lmap[y][x] = MapTile('|')
			tile = self.rng.choice(['*', '&', '"'])
			lmap[y1 + 3][x1 + 3] = MapTile(tile)
		elif w == 9 and h == 9:
			self.furnish_room(x1 + 0, y1 + 0, x1 + 3, y1 + 3)
			self.furnish_room(x1 + 5, y1 + 0, x1 + 8, y1 + 3)
			self.furnish_room(x1 + 0, y1 + 5, x1 + 3, y1 + 8)
			self.furnish_room(x1 + 5, y1 + 5, x1 + 8, y1 + 8)
			tile = self.rng.choice(['*', '&', '"'])
			lmap[y1 + 4][x1 + 4] = MapTile(tile)
		elif w == 11 and h == 11:
			for y in range(y1 + 1, y2, 2):
				lmap[y][x1 + 1] = MapTile('|')
				lmap[y][x2 - 1] = MapTile('|')
			for x in range(x1 + 1, x2, 2):
				lmap[y1 + 1][x] = MapTile('|')
				lmap[y2 - 1][x] = MapTile('|')
			self.furnish_room(x1 + 2, y1 + 2, x1 + 5, y1 + 5)
			self.furnish_room(x1 + 5, y1 + 2, x1 + 8, y1 + 5)
			self.furnish_room(x1 + 2, y1 + 5, x1 + 5, y1 + 8)
			self.furnish_room(x1 + 5, y1 + 5, x1 + 8, y1 + 8)
		elif w > 4 and h > 4:
			lmap[y1 + 1][x1 + 1] = MapTile('#')
			lmap[y1 + 1][x2 - 1] = MapTile('#')
			lmap[y2 - 1][x1 + 1] = MapTile('#')
			lmap[y2 - 1][x2 - 1] = MapTile('#')
			self.place_items(x1 + 2, y1 + 2, x2 - 2, y2 - 2)
		else:
			self.place_items(x1, y1, x2, y2)

	def place_items(self, x1, y1, x2, y2):
		if self.level.depth not in levels:
			return
		
		for y in range(y1 + 1, y2):
			for x in range(x1 + 1, x2):
				chance = self.rng.randint(1, 100)
				if chance <= 20:
					self.place_one_item(x, y, chance)
	
	def place_one_item(self, x, y, chance):
		depth = self.level.depth

		if chance % 5 == 4:
			kind = self.pick_thing("coins")
			if kind != None:
				thing = Item(coins[kind])
				thing.amount = (depth - depth // 2) * 10
				thing.move_to(x, y, self.level)
		elif chance % 5 == 3:
			kind = self.pick_thing("weapons")
			if kind != None:
				thing = Item(weapons[kind])
				thing.move_to(x, y, self.level)
		elif chance % 5 == 2:
			kind = self.pick_thing("spells")
			if kind != None:
				thing = Item(spells[kind])
				thing.move_to(x, y, self.level)
		elif chance % 5 == 1:
			kind = self.pick_thing("armors")
			if kind != None:
				thing = Item(armors[kind])
				thing.move_to(x, y, self.level)
		elif chance % 5 == 0:
			kind = self.pick_thing("foods")
			if kind != None:
				thing = Item(foods[kind])
				thing.move_to(x, y, self.level)
		
	def pick_thing(self, category):
		return self.rng.choice(levels[self.level.depth][category])
	
def render_level(level):
	for y in range(level.height):
		for x in range(level.width):
			tile = level.level_map[y][x]
			tile.icon_id = game_map.make_tile_icon(
				tile.char, x, y)

def update_view(player):
	level = player.level
	lmap = level.level_map
	for y in range(level.height):
		for x in range(level.width):
			tile = lmap[y][x]
			if tile.icon_id < 1:
				pass
			elif tile.seen:
				game_map.fade_tile(tile.icon_id)
			else:
				game_map.hide_tile(tile.icon_id)
	for i in player.in_rooms:
		for y in range(i.y1, i.y2 + 1):
			for x in range(i.x1, i.x2 + 1):
				tile = lmap[y][x]
				game_map.show_tile(tile.icon_id)
	
	game_map.canvas.delete("thing")
	for i in player.level.content:
		for room in player.in_rooms:
			if room.contains(i.x, i.y):
				i.icon_id = game_map.make_thing_icon(
					i.char, i.x, i.y, i.color)
				break
	game_map.canvas.lift(player.icon_id)

def check_death(player):
	if player.wounds >= player.health():
		viewport.delete("all")
		viewport.create_text(
			480, 300, text=defeat_msg, font="Times 18 bold",
			fill="indianred", width=600, justify="left")
		return True
	elif player.victorious:
		viewport.delete("all")
		viewport.create_text(
			480, 300, text=victory_msg, font="Times 18 bold",
			fill="gold", width=600, justify="left")
		return True
	else:
		return False

def handle_walk(dungeon, heading):
	p = dungeon.player
	if check_death(p):
		return
	p.walk(heading)
	dungeon.update()
	update_view(p)
	update_charsheet(p)
	game_hud["ne"] = "Moves: " + str(p.moves)
	game_hud["se"] = "Wounds: %d / %d" % (p.wounds, p.health())
	game_hud["s"] = " / ".join(p.effects.keys())
	game_hud["sw"] = "%d / %d / %d" % (
		p.attack_bonus(),
		p.defense_bonus(),
		p.damage_bonus())
	if p.blessings > 0:
		game_hud["n"] = "blessed"
	elif p.curses > 0:
		game_hud["n"] = "cursed"
	else:
		game_hud["n"] = ""

def handle_ascent(dungeon):
	p = dungeon.player
	if check_death(p):
		return
	p.ascend()
	dungeon.update()
	game_map.clear()
	render_level(p.level)
	update_view(p)
	update_charsheet(p)
	game_hud.lift()
	game_hud["nw"] = "Level " + str(p.level.depth)
	game_hud["ne"] = "Moves: " + str(p.moves)
	game_hud["se"] = "Wounds: %d / %d" % (p.wounds, p.health())

def handle_descent(dungeon):
	p = dungeon.player
	if check_death(p):
		return
	p.descend()
	dungeon.update()
	game_map.clear()
	render_level(p.level)
	update_view(p)
	update_charsheet(p)
	game_hud.lift()
	game_hud["nw"] = "Level " + str(p.level.depth)
	game_hud["ne"] = "Moves: " + str(p.moves)
	game_hud["se"] = "Wounds: %d / %d" % (p.wounds, p.health())

def handle_climb(dungeon):
	p = dungeon.player
	tile = p.level.level_map[p.y][p.x].char
	if tile == '>' or tile == '+':
		handle_ascent(dungeon)
	elif tile == '<':
		handle_descent(dungeon)
	else:
		p.log("There are no stairs here.")

def handle_get(player):
	if check_death(player):
		return
	player.get_item_here()
	update_view(player)
	update_inventory(player)

def handle_spellcast(player):
	if check_death(player):
		return
	player.cast_spell()
	update_charsheet(player)
	game_hud["s"] = " / ".join(player.effects.keys())
	game_hud["sw"] = "%d / %d / %d" % (
		player.attack_bonus(),
		player.defense_bonus(),
		player.damage_bonus())
	game_hud["se"] = "Wounds: %d / %d" % (
		player.wounds, player.health())
	if player.blessings > 0:
		game_hud["n"] = "blessed"
	elif player.curses > 0:
		game_hud["n"] = "cursed"
	else:
		game_hud["n"] = ""

def handle_wield(player):
	if check_death(player):
		return
	
	sel = item_display.selection()
	
	if len(sel) == 0:
		player.log("No items selected.")
		return
	
	player.equip(item_display.index(sel[0]))

	update_charsheet(player)
	update_inventory(player)
	game_hud["sw"] = "%d / %d / %d" % (
		player.attack_bonus(),
		player.defense_bonus(),
		player.damage_bonus())

def handle_eat(player):
	if check_death(player):
		return
	
	sel = item_display.selection()
	
	if len(sel) == 0:
		player.log("No items selected.")
		return
	
	player.eat(item_display.index(sel[0]))

	update_charsheet(player)
	update_inventory(player)
	game_hud["s"] = " / ".join(player.effects.keys())
	game_hud["se"] = "Wounds: %d / %d" % (
		player.wounds, player.health())

def handle_drop(player):
	if check_death(player):
		return
	
	sel = item_display.selection()
	
	if len(sel) == 0:
		player.log("No items selected.")
		return
	
	idx = item_display.index(sel[0])
	item = player.content[idx]

	if item.amount > 1:
		amount = askinteger("Drop item",
			"How many %s to drop?" % (item.a_name(),),
			parent=inventory, initialvalue=item.amount,
			minvalue=1, maxvalue=item.amount)
		if amount == None:
			return
	else:
		amount = 1

	dropped = player.drop(idx, amount)
	dropped.move_to(player.x, player.y, player.level)
	if amount > 1:
		player.log("You drop %s times %d." % (item.a_name(), amount))
	else:
		player.log("You drop %s." % (item.the_name(),))
	update_inventory(player)
	update_view(player)

def handle_action(dungeon):
	p = dungeon.player
	
	if check_death(p):
		return
	elif p.level.any_items_at(p.x, p.y):
		p.get_item_here()
		update_view(p)
		update_inventory(p)
		return

	tile = p.level.level_map[p.y][p.x].char

	if tile == '>' or tile == '+':
		handle_ascent(dungeon)
	elif tile == '<':
		handle_descent(dungeon)
	else:
		p.log("Nothing to do.")

def handle_look(player, x, y):
	if not check_death(player):
		player.look_at(x // scale, y // scale)

def attr_upgrader(player, attr):
	def callback():
		player.__dict__[attr] += 1
		attr_display[attr].configure(
			text=player.__dict__[attr])
		player.log("You now have %d dice in %s!" % (
			player.__dict__[attr], attr))
		player.unused_xp -= xp_attr_cost
		xp_display["Unused"].configure(text=player.unused_xp)
		if player.unused_xp < xp_attr_cost:
			for i in attr_buttons:
				attr_buttons[i].configure(
					state="disabled")
	return callback

def update_charsheet(player):
	for i in attr_display:
		attr_display[i].configure(
			text=player.__dict__[i])
	if player.unused_xp >= xp_attr_cost:
		for i in attr_buttons:
			attr_buttons[i].configure(state="normal")
	
	xp_display["Total"].configure(text=player.total_xp)
	xp_display["Unused"].configure(text=player.unused_xp)

	lbl_wounds.configure(
		text="Wounds: %d/%d" % (player.wounds, player.health()))
	if player.blessings > 0:
		lbl_fortune.configure(text="blessed")
	elif player.curses > 0:
		lbl_fortune.configure(text="cursed")
	else:
		lbl_fortune.configure(text="")

	gear_display["Weapon"].configure(text=str(player.weapon))
	gear_display["Armor"].configure(text=str(player.armor))
	gear_display["Spell"].configure(text=str(player.spell))

	bonus_display["Attack"].configure(text=str(player.attack_bonus()))
	bonus_display["Defense"].configure(text=str(player.defense_bonus()))
	bonus_display["Damage"].configure(text=str(player.damage_bonus()))

def update_inventory(player):
	item_display.delete(*item_display.get_children())
	for i in player.content:
		item_display.insert(
			"", "end", values=(
				i.name, i.amount, i. weight * i.amount,
				i.element if i.element != None else ""))
	items = item_display.get_children()
	if len(items) > 0:
		item_display.selection_set(items[-1])
		item_display.focus(items[-1])
		item_display.focus_set()
	coin = "Gold: %d Silver: %d Amber: %d" % (
		player.coin["gold"],
		player.coin["silver"],
		player.coin["amber"])
	lbl_coin.configure(text=coin)

def bind_commands(dungeon):
	p = dungeon.player
	
	viewport.bind("<1>", lambda event:handle_look(p, event.x, event.y))
	
	all_buttons["Take"].configure(command=lambda: handle_get(p))
	all_buttons["Magic"].configure(command=lambda: handle_spellcast(p))

	all_buttons["Left"].configure(
		command=lambda: handle_walk(dungeon, "w"))
	all_buttons["Down"].configure(
		command=lambda: handle_walk(dungeon, "s"))
	all_buttons["Up"].configure(
		command=lambda: handle_walk(dungeon, "n"))
	all_buttons["Right"].configure(
		command=lambda: handle_walk(dungeon, "e"))
	all_buttons["Climb"].configure(
		command=lambda: handle_climb(dungeon))

	top.bind('g', lambda event: handle_get(p))
	top.bind('m', lambda event: handle_spellcast(p))

	top.bind('h', lambda event: handle_walk(dungeon, "w"))
	top.bind('j', lambda event: handle_walk(dungeon, "s"))
	top.bind('k', lambda event: handle_walk(dungeon, "n"))
	top.bind('l', lambda event: handle_walk(dungeon, "e"))

	top.bind("<Key-Left>", lambda event: handle_walk(dungeon, "w"))
	top.bind("<Key-Down>", lambda event: handle_walk(dungeon, "s"))
	top.bind("<Key-Up>", lambda event: handle_walk(dungeon, "n"))
	top.bind("<Key-Right>", lambda event: handle_walk(dungeon, "e"))

	top.bind('<Key-less>', lambda event: handle_descent(dungeon))
	top.bind('<Key-greater>', lambda event: handle_ascent(dungeon))

	top.bind('<Key-space>', lambda event: handle_action(dungeon))

	charsheet.bind("<Map>", lambda event: update_charsheet(p))
	inventory.bind("<Map>", lambda event: update_inventory(p))

	for i in attr_buttons:
		attr_buttons[i].configure(
			command=attr_upgrader(p, i))
	
	inventory.bind("w", lambda event: handle_wield(p))
	inventory.bind("e", lambda event: handle_eat(p))
	inventory.bind("d", lambda event: handle_drop(p))

	inv_buttons["Wield/wear"].configure(
		command=lambda: handle_wield(p))
	inv_buttons["Eat"].configure(
		command=lambda: handle_eat(p))
	inv_buttons["Drop"].configure(
		command=lambda: handle_drop(p))

dungeon_gen = PalaceFloorGenerator(random.Random())
game_dungeon = None
game_map = None
game_hud = None

def new_game():
	global game_dungeon, game_map, game_hud
	
	answer = askinteger("New game", "Enter game number:",
		parent=top, initialvalue=random.getrandbits(32))
	if answer == None:
		return
	else:
		dungeon_gen.rng.seed(answer)
	
	game_dungeon = Dungeon(dungeon_gen)
	game_dungeon.player.name = "Hero"
	game_dungeon.player.log = log_to_console
	game_dungeon.update()

	if game_map == None:
		game_map = MapView(viewport)

	viewport.delete("all")
	render_level(game_dungeon.level1)
	update_view(game_dungeon.player)

	# Clearing the viewport also removes the previous HUD, if any.
	game_hud = HUD(viewport)
	game_hud["nw"] = "Level 1"
	game_hud["ne"] = "Moves: 0"
	game_hud["se"] = "Wounds: 0 / 24"
	game_hud["sw"] = "%d / %d / %d" % (
		game_dungeon.player.attack_bonus(),
		game_dungeon.player.defense_bonus(),
		game_dungeon.player.damage_bonus())

	console.delete("1.0", "end")
	log_to_console("Game #%d started at %s."
		% (answer, time.strftime("%H:%M")))
	print("Game #%d started at %s."
		% (answer, time.strftime("%H:%M")))
	log_to_console(
		"Work in progress. Walk around, bump into things, have fun.")

	bind_commands(game_dungeon)

game_menu.entryconfigure(0, command=new_game)
top.bind("<Control-n>", lambda event: new_game())

print("Stats: %d foods, %d weapons, %d armors, %d spells, %d creatures." %
	(len(foods), len(weapons), len(armors), len(spells), len(creatures)))

if len(sys.argv) > 1:
	if sys.argv[1] == "-q" or sys.argv[1] == "-quiet":
		quiet = True

render_title_screen(viewport)
top.mainloop()
